﻿using RecrownedAthenaeum.Graphics.Render;
using System;
using OpenTK;
using OpenTK.Graphics;

namespace RecrownedAthenaeum.Graphics
{
    /// <summary>
    /// An object that represents a ninepatch.
    /// </summary>
    public class NinePatch : ISpecialDrawable
    {
        /// <summary>
        /// Dimensions in ninepatch. May also represent position in texture atlas.
        /// </summary>
        public Rectangle textureRegion;
        readonly Texture2D texture;
        readonly int left, right, bottom, top;

        Rectangle[] sourcePatches;

        /// <summary>
        /// A nine patch object.
        /// </summary>
        /// <param name="texture">Texture used for the nine patch. Dimensions must be greater than their sum border counter parts. If used as part of texture atlas, the texture should be the texture of the entire atlas.</param>
        /// <param name="left">Left side.</param>
        /// <param name="right">Right side.</param>
        /// <param name="bottom">Bottom side.</param>
        /// <param name="top">Top side.</param>
        /// <param name="textureBounds">The dimensions and potentially the coordinates of the rectangle on an atlas. If left to default of null, will only be set to texture bounds.</param>
        public NinePatch(Texture2D texture, int left, int right, int bottom, int top, Rectangle? textureBounds = null)
        {
            this.texture = texture;
            if (textureBounds.HasValue) textureRegion = textureBounds.Value; else textureRegion = texture.Bounds;
            if (left + right >= textureRegion.Width) throw new ArgumentOutOfRangeException("left and right values cannot be greater than or equal to the width of region. Left value is " + left + " and right value is " + right + ". Bounds of texture are: " + textureRegion);
            if (bottom + top >= textureRegion.Height) throw new ArgumentOutOfRangeException("Bottom and top values cannot be greater than or equal to the width of region. Bottom value is " + bottom + " and top value is " + top + ".");
            this.left = left;
            this.right = right;
            this.bottom = bottom;
            this.top = top;

            Rectangle[] patches =
{
                new Rectangle(0, 0, left, bottom),
                new Rectangle(left, 0, textureRegion.Width - left - right, bottom),
                new Rectangle(textureRegion.Width - right, 0, right, bottom),
                new Rectangle(0, bottom, left, textureRegion.Height - top - bottom),
                new Rectangle(left, bottom, textureRegion.Width - left - right, textureRegion.Height - bottom - top),
                new Rectangle(textureRegion.Width - right, bottom, right, textureRegion.Height - top - bottom),
                new Rectangle(0, textureRegion.Height - top, left, top),
                new Rectangle(left, textureRegion.Height - top, textureRegion.Width - left - right, top),
                new Rectangle(textureRegion.Width - right, textureRegion.Height - top, right, top),
            };

            for (int i = 0; i < patches.Length; i++)
            {
                patches[i].X += textureRegion.X;
                patches[i].Y += textureRegion.Y;
            }

            sourcePatches = patches;

        }

        private Rectangle[] GenenerateDestinationRectangles(int width, int height)
        {
            Rectangle[] patches =
            {
                new Rectangle(0, 0, left, bottom),
                new Rectangle(left, 0, width - left - right, bottom),
                new Rectangle(width -right, 0, right, bottom),
                new Rectangle(0, bottom, left, height - bottom - top),
                new Rectangle(left, bottom, width - left - right, height - top - bottom),
                new Rectangle(width - right, bottom, right, height - top - bottom),
                new Rectangle(0, height - top, left, top),
                new Rectangle(left, height - top, width - left - right, top),
                new Rectangle(width - right, height - top, right, top),
            };
            return patches;
        }

        /// <summary>
        /// Draws the ninepatch.
        /// </summary>
        /// <param name="spriteBatch">Batch to use.</param>
        /// <param name="color">The color of the patch.</param>
        /// <param name="destination">Where to the patch.</param>
        public void Draw(ConsistentSpriteBatch spriteBatch, Color4 color, Rectangle destination)
        {
            try
            {
                spriteBatch.End();
            } catch (InvalidOperationException)
            {
                throw new InvalidOperationException("Begin must be called to draw a nine patch!");
            }

            spriteBatch.BeginWithoutSaving(samplerState: SamplerState.PointClamp);

            Rectangle[] destinations = GenenerateDestinationRectangles(destination.Width, destination.Height);
            for (int i = 0; i < destinations.Length; i++)
            {
                destinations[i].X += destination.X;
                destinations[i].Y += destination.Y;
                spriteBatch.Draw(texture, destinations[i], sourcePatches[i], color);
            }
            spriteBatch.End();

            spriteBatch.Begin();
        }

        /// <summary>
        /// </summary>
        /// <param name="spriteBatch">Spritebatch to use.</param>
        /// <param name="destination">The destination to draw the patch.</param>
        /// <param name="color">The tint for each patch.</param>
        /// <param name="rotation">Not considered for 9patches.</param>
        /// <param name="origin">Not considered for 9patches.</param>
        public void Draw(ConsistentSpriteBatch spriteBatch, Rectangle destination, Color4 color, float rotation = 0, Vector2 origin = default(Vector2))
        {
            if (rotation != 0) throw new NotImplementedException("Ninepatches can't be rotated.");
            if (origin != default(Vector2))
            {
                destination.X -= (int)origin.X;
                destination.Y -= (int)origin.Y;
            }
            Draw(spriteBatch, color, destination);
        }
    }
}
